﻿using System;
using System.Threading;
using System.Threading.Tasks;
using GrapeCity.CalcEngine;

namespace SpreadNETZetaCalculator
{
    /// <summary>
    /// Class implementing the custom fuction to calculate the Riemann ZETA function for a set of points along a vertical line (e.g. the critical line x = 1/2).
    /// </summary>
    /// <remarks>
    /// This implementation calculates the analytic continuation ζ(s) in the complex plane, which is defined for s > 0.
    /// Usage of this function is as follows:
    /// ZETA(re, im, count, incr, iter) where:
    /// re = the real component of the initial point to calculate (must be > 0)
    /// im = the imaginary component of the initial point to calculate (values are symetric about x-axis => ζ(x+bi) = ζ(x-bi) )
    /// count = the number of points to calculate
    /// incr = the increment between each point to calculate
    /// iter = the number of iterations to calculate -- the higher this number, the more accurate the calculation (and longer it will take)
    /// </remarks>
    internal sealed class ZetaFunction : Function
    {   // keep count of function calls
        internal static int cnt = 0;
        // time calculation
        internal static DateTime t1;
        internal static DateTime t2;
        /// <summary>
        /// Create the ZetaFunction.
        /// </summary>
        public ZetaFunction() : base("ZETA", 4, 5, FunctionAttributes.Array)
        { // minimum args 4, maximum args 5, returns array value
        }
        /// <summary>
        /// Evaluate the ZETA function.
        /// </summary>
        /// <param name="args">IArguments specifying arguments passed to ZETA</param>
        /// <param name="result">IValue for returning calculation result</param>
        protected override void Evaluate(IArguments args, IValue result)
        {
            cnt++; // increment function call counter
            t1 = DateTime.Now; // start timing evaluation
            long l = 10000L; // default to 10K iterations if not specified
            // get function arguments
            double d = args[0].GetPrimitiveValue(args.EvaluationContext).GetNumber();  // real part
            double d1 = args[1].GetPrimitiveValue(args.EvaluationContext).GetNumber(); // imaginary part
            long count = (long)args[2].GetPrimitiveValue(args.EvaluationContext).GetNumber(); // number of points to calculate
            double incr = args[3].GetPrimitiveValue(args.EvaluationContext).GetNumber(); // increment between points
            if (args.Count > 4) // iterations is optional
                l = (long)args[4].GetPrimitiveValue(args.EvaluationContext).GetNumber();
            IValuesArray array = result.CreateValuesArray((int)count, 2); // create array to return values
            double[,] arr = new double[count, 2]; // create local array to calculate values (can't set values to IValuesArray from another thread)
            // do the ZETA calculation in a Task using Parallel.For for better performance
            Task t = new Task(delegate
            {
                Parallel.For(0, count, i =>
                    {
                    var d2 = 0.0;
                    var d3 = 0.0;
                    var d4 = 0.0;
                    var d5 = 0.0;
                    var dd1 = d1 + i * incr;
                    for (var l1 = 1; l1 <= l; l1++)
                    { // calculate Dirichlet Eta series partial sum of l iterations
                        var d6 = 2 * l1 - 1;
                        var d7 = 2 * l1;
                        var d14 = 0.0;
                        var d15 = 0.0;
                        var d16 = 0.0;
                        var d8 = (dd1 * Math.Log(d6)) % 6.2831853071795862;
                        var d9 = (dd1 * Math.Log(d7)) % 6.2831853071795862;
                        var d10 = -d;
                        var d11 = Math.Pow(d6, d10);
                        var d12 = Math.Pow(d7, d10);
                        d2 += d11 * Math.Cos(d8) - d12 * Math.Cos(d9);
                        d3 += d12 * Math.Sin(d9) - d11 * Math.Sin(d8);
                        d4 += Math.Pow(d6, -0.5) * Math.Cos(d8) - Math.Pow(d7, -0.5) * Math.Cos(d9);
                        d14 = Math.Pow(d7, -0.5) * Math.Sin(d9);
                        d15 = Math.Pow(d6, -0.5);
                        d16 = Math.Sin(d8);
                        d5 += d14 - d15 * d16;
                    }
                    // calculate Zeta from Eta value
                    var d17 = 1.0 - Math.Pow(2, 1.0 - d) * Math.Cos(d1 * Math.Log(2));
                    var d18 = Math.Pow(2, 1.0 - d) * Math.Sin(d1 * Math.Log(2));
                    var d19 = (d17 * d2 + d18 * d3) / (Math.Pow(d17, 2) + Math.Pow(d18, 2));
                    var d20 = (d17 * d3 - d18 * d2) / (Math.Pow(d17, 2) + Math.Pow(d18, 2));
                    Interlocked.Exchange(ref arr[i, 0], d19);
                    Interlocked.Exchange(ref arr[i, 1], d20);
                });
            });
            t.RunSynchronously();  // need to run this task synchronously on the UI thread
            // copy values from arr to the IValuesArray
            for (int i = 0; i < count; i++)
            {
                array[i, 0].SetValue(arr[i, 0]);
                array[i, 1].SetValue(arr[i, 1]);
            }
            result.SetValue(array); // set result -- this is important! If you don't return a result, then the cells remain marked dirty and Evaluate will be called again...
            t2 = DateTime.Now; // time calculation
        }
    }
}
