﻿using System;
using System.Threading;
using System.Threading.Tasks;
using FarPoint.CalcEngine;

namespace SpreadNETZetaCalculator
{
    /// <summary>
    /// Class implementing the custom fuction to calculate the Riemann ZETA function for a set of points along a vertical line (e.g. the critical line x = 1/2).
    /// </summary>
    /// <remarks>
    /// This implementation uses the old FarPoint.CalcEngine.FunctionInfo type, which is obsolete but still supported for backwards compatibility.
    /// This implementation calculates the analytic continuation ζ(s) in the complex plane, which is defined for s > 0.
    /// Usage of this function is as follows:
    /// ZETA(re, im, count, incr, iter) where:
    /// re = the real component of the initial point to calculate
    /// im = the imaginary component of the initial point to calculate
    /// count = the number of points to calculate
    /// incr = the increment between each point to calculate
    /// iter = the number of iterations to calculate -- the higher this number, the more accurate the calculation (and longer it will take)
    /// </remarks>
    internal sealed class ZetaFunctionInfo : FunctionInfo
    {   // keep count of function calls
        internal static int cnt = 0;
        // time calculation
        internal static DateTime t1;
        internal static DateTime t2;
        // required overrides for FunctionInfo
        public override string Name => "ZETA";
        public override int MinArgs => 4;
        public override int MaxArgs => 5;
        /// <summary>
        /// Evaluate the ZETA function.
        /// </summary>
        /// <param name="args">Array of object containing function arguments</param>
        /// <returns>ArrayResult object containing calculated results</returns>
        public override object Evaluate(object[] args)
        {
            cnt++;
            t1 = DateTime.Now;
            int l = 10000;
            double d = CalcConvert.ToDouble(args[0]);  // real part
            double d1 = CalcConvert.ToDouble(args[1]); // imaginary part
            int count = CalcConvert.ToInt(args[2]); // number of points to calculate
            double incr = CalcConvert.ToDouble(args[3]); // increment between points
            if (args.Length > 4)
                l = CalcConvert.ToInt(args[4]); // iterations
            double[,] arr = new double[count, 2]; // create array to return values
            // do the ZETA calculation in a Task using Parallel.For for better performance
            Task t = new Task(delegate
            {
                Parallel.For(0, count, i =>
                {
                    var d2 = 0.0;
                    var d3 = 0.0;
                    var d4 = 0.0;
                    var d5 = 0.0;
                    var dd1 = d1 + i * incr;
                    for (var l1 = 1; l1 <= l; l1++)
                    {
                        var d6 = 2 * l1 - 1;
                        var d7 = 2 * l1;
                        var d14 = 0.0;
                        var d15 = 0.0;
                        var d16 = 0.0;
                        var d8 = (dd1 * Math.Log(d6)) % 6.2831853071795862;
                        var d9 = (dd1 * Math.Log(d7)) % 6.2831853071795862;
                        var d10 = -1 * d;
                        var d11 = -1 * 0.5;
                        var d12 = Math.Pow(d6, d10);
                        var d13 = Math.Pow(d7, d10);
                        d2 += d12 * Math.Cos(d8) - d13 * Math.Cos(d9);
                        d3 += d13 * Math.Sin(d9) - d12 * Math.Sin(d8);
                        d4 += Math.Pow(d6, d11) * Math.Cos(d8) - Math.Pow(d7, d11) * Math.Cos(d9);
                        d14 = Math.Pow(d7, d11) * Math.Sin(d9);
                        d15 = Math.Pow(d6, d11);
                        d16 = Math.Sin(d8);
                        d5 += d14 - d15 * d16;
                    }
                    var d17 = 1.0 - Math.Pow(2, 1.0 - d) * Math.Cos(d1 * Math.Log(2));
                    var d18 = Math.Pow(2, 1.0 - d) * Math.Sin(d1 * Math.Log(2));
                    var d19 = (d17 * d2 + d18 * d3) / (Math.Pow(d17, 2) + Math.Pow(d18, 2));
                    var d20 = (d17 * d3 - d18 * d2) / (Math.Pow(d17, 2) + Math.Pow(d18, 2));
                    Interlocked.Exchange(ref arr[i, 0], d19);
                    Interlocked.Exchange(ref arr[i, 1], d20);
                });
            });
            t.RunSynchronously(); // need to run this task synchronously on the UI thread
            t2 = DateTime.Now;
            return new ArrayResult(arr); // return as ArrayResult (see below)
        }
    }
    /// <summary>
    /// Class inheriting from FarPoint.CalcEngine.CalcArray used to return the array result.
    /// </summary>
    /// <remarks>
    /// The old calculation engine (FarPoint.CalcEngine.dll) required implemnting a class to return an array of values.
    /// The CalcArray abstract class provides the base type for implementing a wrapper around the array to return.
    /// </remarks>
    public class ArrayResult : CalcArray
    {
        double[,] values;
        public ArrayResult(double[,] values)
        {
            if (values == null)
                values = new double[0, 0];
            this.values = values;
        }
        public override int RowCount => values.GetUpperBound(0) - values.GetLowerBound(0) + 1;
        public override int ColumnCount => values.GetUpperBound(1) - values.GetLowerBound(1) + 1;
        public override object GetValue(int row, int column)
        {
            return values[row, column];
        }
    }
}
